Require Import Setoid.
Require Import ClassicalRealizability.


Definition redeq := fun p p' => p = p' \/ red p p'.
Definition dot (π : Π) : formula := fun π' => π' = π.

Hypothesis non_conf : forall t π, t·π <> π.
Hypothesis non_conf2 : forall t u π, t·u·π <> π.

Lemma redeq_red : forall p p', p <> p' -> redeq p p' -> red p p'.
Proof. intros p p' Hp []; intuition. Qed.

Theorem specification_1 : forall t, (t ⊩ one) -> (forall u π, red (t ★ u·π) (u ★ π)).
Proof.
intros t Ht u π.
assert (Hpole : forall p, p ∈ ⫫ <-> redeq p (u ★ π)). admit. (* we choose the pole *)
apply redeq_red. intro H. inversion H. eelim non_conf. eassumption.
rewrite <- Hpole. apply Ht. exists (dot π). Ksolve.
Ksolve. compute in Hπ. subst. rewrite Hpole. now left.
Qed.

Theorem specification_bot_impl_bot : forall t, (t ⊩ ⊥ → ⊥) -> (forall u π, exists π', red (t ★ u·π) (u ★ π')).
Proof.
intros t Ht u π.
assert (Hpole : forall p, p ∈ ⫫ <-> exists π', redeq p (u ★ π')). admit. (* we choose the pole *)
(*assert (u ⊩ ⊥).
  start. rewrite Hpole. exists π0. now left.*)
assert (t ★ u·π ∈ ⫫) as Hbot.
  Ksolve. start. rewrite Hpole. exists π. now left.
rewrite Hpole in Hbot. destruct Hbot as [π' Hbot].
exists π'. destruct Hbot.
  admit. (* requires 2 different u to prove that t cannot be equal to both *)
  assumption.
Qed.

Theorem specification_Bool : forall t, (t ⊩ ∀Z, Z → Z → Z) ->
  (forall u v π, red (t ★ u·v·π) (u ★ π) \/ red (t ★ u·v·π) (v ★ π)).
Proof.
intros t Ht u v π.
assert (Hpole : forall p, p ∈ ⫫ <-> redeq p (u ★ π) \/ redeq p (v ★ π)). admit. (* we choose the pole *)
setoid_replace (red (t ★ u·v·π) (u ★ π)) with (redeq (t ★ u·v·π) (u ★ π)).
setoid_replace (red (t ★ u·v·π) (v ★ π)) with (redeq (t ★ u·v·π) (v ★ π)).
rewrite <- Hpole. apply Ht. exists (dot π). Ksolve.
Ksolve. compute in Hπ. subst. rewrite Hpole. now do 2 left.
Ksolve. compute in Hπ. subst. rewrite Hpole. now right; left.
unfold redeq. intuition. inversion H0. eelim non_conf2. eassumption.
unfold redeq. intuition. inversion H0. eelim non_conf2. eassumption.
Qed.

